"""
This file contains some helper functions and stuff
"""
import sys
import time
from datetime import datetime
import os
import numpy as np


class bcolors:
    """
    This class contains colors for the terminal
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Timer:
    """
    Basic timer class to measure different processes easily
    """

    # An internal storage for different processes to measure
    processes = None

    def __init__(self):
        """
        Initializes timer.
        """
        # Saves all processes currently in use.
        self.processes = {}

    def start(self, process):
        """
        Starts a process specific timer
        """
        assert isinstance(process, str)
        self.processes[process] = time.time()

    def get_time(self, process, reset: bool = True):
        """
        Returns the current time elapsed and resets timer

        :param reset: if true, timer is resetted
        :param process: the process of which to obtain the timer from
        """
        if reset:
            start_time = self.processes.pop(process, None)
        else:
            start_time = self.processes.get(process, None)
        if start_time is None:
            raise KeyError('Process not found.')
        else:
            return np.round(time.time() - start_time, 2)

    def reset(self, process):
        """
        Reset timer of a process.

        :param process: the process of which to reset the timer from
        """
        self.processes[process] = None

    def reset_all(self):
        """
        Resets all timers
        """
        self.processes = {}


class Logger:
    """
    Class that logs messages.
    """

    # Stores the path of the log file
    path = ""

    # Attribute stores whether there is a progressbar currently
    status_bar_active = False
    last_status_bar = None

    # Whether or whether not the logger is initialized
    initialized = False

    @staticmethod
    def initialize(log_path=None, **kwargs):
        """Initialize method

        :param log_path: the name of the log
        """

        # Save variables
        if log_path is not None:
            # Define path
            Logger.path = os.path.join(log_path, 'log.txt')
            # Create file if it does not exist yet
            if not os.path.isfile(Logger.path):
                open(Logger.path, 'w+')
        else:
            Logger.path = None

        # Set initialized
        Logger.initialized = True

    @staticmethod
    def clear():
        """
        Clears the log file

        :return:
        """

        # Clears the file
        with open(Logger.path, 'w') as file:
            file.write("")

    @staticmethod
    def log(message: str, type: str = "", in_cli: bool = True, new_line: bool = False):
        """
        Writes a log message to log file. Message will also be printed in terminal.

        :param message: message to be logged
        :param type: the type of the message
        :param in_cli: whether to also print the message
        """

        # Check if content shall be in a new line
        prefix = "\n" if new_line else ""

        if in_cli:
            Logger.out(message, type, new_line)

        if Logger.initialized:
            if Logger.path is not None:
                with open(Logger.path, 'a+') as file:
                    if message != '':
                        time_str = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
                        log_message = Logger._get_content(message, type, time_str)
                        file.write(prefix + log_message)
                    file.write('\n')

    @staticmethod
    def out(message: str, type: str, new_line: bool = False):
        """
        Only writes message to terminal for debugging purposes.

        :param message: message to be logged
        :param type: the type of the message
        """

        # Get the CLI message to display
        cli_message = Logger._get_content(message, type)

        # Check for new line prefix
        prefix = "\n" if new_line else ""

        if Logger.status_bar_active:
            sys.stdout.write("\r" + prefix + cli_message)
            sys.stdout.flush()
            print("")
            if Logger.last_status_bar is not None:
                sys.stdout.write(Logger.last_status_bar)
                sys.stdout.flush()
        else:
            print(prefix + cli_message)

    @staticmethod
    def print_status_bar(title: str = "done", done: float = 0, bar_width: int = 25):
        """
        This function shows a "running" status bar to visualize progress

        :param title: the title that shall be displayed besides the bar
        :param done: the percentage of the bar
        :param bar_width: how wide the bar shall be
        """

        # Check if done is weird
        if done > 100:
            done = 100

        # The start buffer is appended before the loading bar
        start_buffer = bcolors.OKBLUE + "RUN     "

        Logger.status_bar_active = True
        if done == 100.0:
            Logger.last_status_bar = "\r" + start_buffer + "100.0% " + ('.' * bar_width) + " " + title + "                    " + bcolors.ENDC
        else:
            status_string = ""
            state = "d"
            for j in range(bar_width):
                nextstate = state
                if int(done) >= ((j / bar_width) * 100):
                    status_string += "."
                else:
                    if state == 'd':
                        nextstate = 'nd'
                        status_string += "."
                        status_string += bcolors.ENDC
                    else:
                        status_string += "."
                state = nextstate
            status_string += " "
            done_string = "{:.2f}".format(done)
            if done < 10:
                done_string = done_string + " "
            Logger.last_status_bar = "\r" + start_buffer + done_string + "% " + status_string + title + "                    " + bcolors.ENDC

        sys.stdout.write(Logger.last_status_bar)
        if done != 100.0:
            sys.stdout.flush()

    @staticmethod
    def end_status_bar():
        """
        This function must be called after ending a status bar progress
        """

        Logger.status_bar_active = False
        print("")

    @staticmethod
    def _get_content(message: str, type: str, log_stamp: str = None):
        """

        :param message: the raw message to log or print
        :param type: the type of the message
        :param log_stamp: when passed, it returns a log string
        :return: a string for cli logging or writing to rtf
        """

        # Check whether the type is known
        if type in ['ERROR', 'WARNING', 'INFO', 'SUCCESS', 'HEADLINE', '']:

            # Create prephase when log stamp is given
            if log_stamp is not None:

                # Create a log message
                if type == 'INFO':
                    log_message = log_stamp + "     INFO    \t" + message
                elif type == 'WARNING':
                    log_message = log_stamp + "     WARN    \t" + message
                elif type == 'ERROR':
                    log_message = "\n" + log_stamp + "     ERROR   \t" + message + "\n"
                elif type == 'SUCCESS':
                    log_message = log_stamp + "     DONE    \t" + message + "\n\n"
                elif type == 'HEADLINE':
                    log_message = log_stamp + "     HEAD    \t" + message + "\n\n"
                else:
                    log_message = log_stamp + "     LOG     \t" + message

                return log_message

            else:

                # Print loading message
                if type == 'INFO':
                    cli_message = bcolors.OKBLUE + "INFO    " + message + bcolors.ENDC
                elif type == 'WARNING':
                    cli_message = bcolors.WARNING + "WARN    " + message + bcolors.ENDC
                elif type == 'ERROR':
                    cli_message = bcolors.FAIL + "ERROR   " + message + bcolors.ENDC
                elif type == 'SUCCESS':
                    cli_message = bcolors.OKGREEN + "DONE    " + message + bcolors.ENDC + "\n\n"
                elif type == 'HEADLINE':
                    cli_message = "\n" + bcolors.UNDERLINE + message + bcolors.ENDC + "\n"
                else:
                    cli_message = "LOG     " + message

                return cli_message

        else:

            # Notify that the output type is unknown
            raise ValueError("ERROR: Logger out function does not recognize the message type")



"""
This file contains metric loggers that can be added in a job configuration. They are called
after each epoch to enable the user to log custom metrics e.g. to WANDB or the filesystem.
"""


class BaseMetricLogger:
    """
    This is the base evaluator class that implements some general methods that every evaluator shall have
    """

    def single_eval_batch_operations(self, epoch, model, inputs, labels, log_callback, job_folder):
        """
        This method is called once for each mini batch in the evaluation dataset. Use this function
        to e.g. compute own averages or KPIs over the eval mini batches. You can use attributes in your
        metrics logger classes to access data.

        :param epoch: the current epoch of the logger being called
        :param model: the current model during training
        :param inputs: the last inputs in the epoch
        :param labels: the last labels in the epoch
        :param log_callback: wandb worker to log metrics there
        :param job_folder: folder where job data / stats / metrics can be logged
        """
        pass

    def log_metrics(self, epoch, model, inputs, labels, log_callback, wandb_worker, job_folder):
        """
        With this method, the user can create own metrics and log them. If you use the wb_worker,
        make sure NOT TO COMMIT the log message. The runner will log the metrics all at once later.
        This method is called AFTER the evaluation loop and can be used to log metrics or work with
        local class data. If you want to perform operations foreach mini batch, use the method
        self.single_eval_batch_operations()

        :param epoch: the current epoch of the logger being called
        :param model: the current model during training
        :param inputs: the last inputs in the epoch
        :param labels: the last labels in the epoch
        :param log_callback: callback function to log stuff
        :param wandb: wandb worker to log metrics there
        :param job_folder: folder where job data / stats / metrics can be logged
        """
        pass



"""
This file contains the multi runner class which can run jobs parallel using threads

TODO: implement
"""


class MultiRunner:
    """
    MultiRunner instance

    TODO: implement the multi runner (split the job queue and use n runner classes for that)
    """

    pass



"""
This file contains different losses that can be set in a job
"""
import torch
import numpy as np
from torch import nn


class DiceCoefficient(nn.Module):
    def __init__(self, eps=1e-4, **params):
        self.eps = eps
        super().__init__()

    def forward(self, inputs, targets, return_per_channel_dsc=False):
        # Compute the dice coefficient

        assert inputs.size() == targets.size(), "Input sizes must be equal."
        assert inputs.dim() == 5, "Input must be a 5D Tensor."
        uniques = np.unique(targets.numpy())
        assert set(list(uniques)) <= set([0, 1]), "target must only contain zeros and ones"

        dsc_per_channel = torch.divide((2 * inputs * targets), (inputs + targets + self.eps)).sum(dim=(0, 3, 2, 4))

        organ_sizes = (targets == 1).sum(dim=(0, 3, 2, 4))
        dsc_per_channel = torch.divide(dsc_per_channel, organ_sizes + self.eps)

        dsc_avg = dsc_per_channel.mean()

        if return_per_channel_dsc:
            return dsc_avg, dsc_per_channel

        return dsc_avg


class DiceLoss(nn.Module):
    def __init__(self, eps=1e-8, **params):
        self.eps = eps
        super().__init__()

    def forward(self, inputs, targets, reduce_method="mean", return_per_channel_dsc=False):
        dice = DiceCoefficient(eps=self.eps)(inputs, targets, return_per_channel_dsc=return_per_channel_dsc)
        if return_per_channel_dsc:
            loss, per_channel = dice
            return 1 - loss, per_channel

        return 1 - dice


class MSELoss(nn.Module):
    """
    Determines l1 loss which is the absolute difference between input and output.
    """

    def __init__(self, **params):
        """
        Initialize method of the MSE Loss object

        :param reduction:   'mean' will determine the mean loss over all elements (across batches) while
                            'sum' will determine the summation of losses over all elements
        """
        super().__init__()
        if 'reduction' in params:
            self.reduction = params['reduction']
        else:
            self.reduction = "mean"

    def forward(self, output_batch, input_batch):

        # Determine loss
        loss = nn.MSELoss(reduction=self.reduction)(output_batch, input_batch)

        # In case of summation we want the batch loss, hence we divide by the batch size
        if self.reduction == "sum":
            loss = loss / input_batch.shape[0]

        return loss


class HingeLoss(nn.Module):

    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(self, inputs, targets):

        # targets[targets == 0] = -1
        # loss_function = torch.nn.HingeEmbeddingLoss(reduction="sum")
        # loss = loss_function(inputs, targets)

        loss_function = torch.nn.BCELoss()
        loss = loss_function(inputs, targets)

        return loss



"""
This file contains the evaluator class which evaluates the trained model after training
or ad hoc.
"""


class BaseEvaluator:
    """
    This is the base evaluator class that implements some general methods that every evaluator shall have
    """

    def evaluate(self, trained_model, evaluation_set, job_path):
        """
        This method is abstract and has to be implemented by

        :param trained_model:
        :param job_path:
        :param evaluation_set:
        :return: the score of evaluation
        """
        pass
